--[[
	Todo

]]

local MAJOR, MINOR = "LibNameplate-1.0", 1
if not LibStub then error(MAJOR .. " requires LibStub.") return end

local lib = LibStub:NewLibrary(MAJOR, MINOR)
if not lib then return end

lib.callbacks = lib.callbacks or LibStub("CallbackHandler-1.0"):New(lib)
if not lib.callbacks then	error(MAJOR .. " CallbackHandler-1.0.") return end

-- Globals
local WorldFrame_			= WorldFrame
local select_				= select
local UnitExists_			= UnitExists
local UnitGUID_				= UnitGUID
local UnitName_				= UnitName
local string_gsub			= string.gsub
local pairs_				= pairs
local table_remove			= table.remove
local table_insert			= table.insert
local GetMouseFocus_		= GetMouseFocus
local GetNumRaidMembers_	= GetNumRaidMembers
local GetNumPartyMembers_	= GetNumPartyMembers
local UnitIsUnit_			= UnitIsUnit
local GetRaidTargetIndex_	= GetRaidTargetIndex
local UnitInParty_			= UnitInParty
local tonumber_				= tonumber

--Conts
local scanDelay		= 1 -- Scan for new nameplates every 1 seconds
local updateDelay	= 1 -- Update nameplate info every 1 seconds.


local DEBUG = false
local function CmdHandler()
	DEBUG = not DEBUG
end
SlashCmdList["LIBNAMEPLATEDEBUG"] = CmdHandler
SLASH_LIBNAMEPLATEDEBUG1 = "/lnbug"
local function debugPrint(...)
	if DEBUG == true then
		print(...)
	end
end

local regionNames = { "threatGlow", "healthBorder", "castBorder", "castNostop", "spellIcon", "highlightTexture", "nameText", "levelText", "dangerSkull", "raidIcon", "eliteIcon" }


--~ local raidIconTexture = "Interface\\TargetingFrame\\UI-RaidTargetingIcons"
--"Star", "Circle", "Diamond", "Triangle", "Moon", "Square", "Cross", "Skull"
--			= left							right
--			= ULx	ULy		LLx		LLy		URx		URy		LRx		LRy
-- star		= 0		0		0		0.25	0.25	0		0.25	0.25
-- circle	= 0.25	0		0.25	0.25	0.5		0		0.5		0.25
-- diamond	= 0.5	0		0.5		0.25	0.75	0		0.75	0.25
-- triangle	= 0.75	0		0.75	0.25	1		0		1		0.25
-- moon		= 0		0.25	0		0.5		0.25	0.25	0.25	0.5
-- square 	= 0.25	0.25	0.25	0.5		0.5		0.25	0.5		0.5
-- cross	= 0.5	0.25	0.5		0.5		0.75	0.25	0.75	0.5
-- skull	= 0.75	0.25	0.75	0.5		1		0.25	1		0.5
local raidIcons = {--from GetTexCoord. input is ULx and ULy (first 2 values).
	[0]		= {
		[0]		= 1, --star 
		[0.25]	= 5, --moon
	},
	[0.25]	= {
		[0]		= 2, --circle 
		[0.25]	= 6, --square
	},
	[0.5]	= {
		[0]		= 3, --star 
		[0.25]	= 7, --cross
	},
	[0.75]	= {
		[0]		= 4, --star 
		[0.25]	= 8, --skull
	},
}

lib.nameplates		= lib.nameplates or {}
lib.GUIDs			= lib.GUIDs or {}
lib.names			= lib.names or {}

local colorToClass = {}
local function pctToInt(number) return math.floor((100*number) + 0.5) end
for classname, color in pairs_(RAID_CLASS_COLORS) do
	colorToClass["C"..pctToInt(color.r)+pctToInt(color.g)+pctToInt(color.b)] = classname
end


local function IsNamePlateFrame(frame)
	if frame:GetName() then
		return false
	end
	local region1, overlayRegion = frame:GetRegions()
	if not overlayRegion then
		return false
	end
	if overlayRegion:GetObjectType() ~= "Texture" then
		return false
	end
	if overlayRegion:GetTexture() ~= "Interface\\Tooltips\\Nameplate-Border" then
		return false
	end
	return true
end

local function threatByColor( object , name)
	local redCan, greenCan, blueCan, alphaCan = object:GetVertexColor()
	local shown
	if not object:IsShown() then return "LOW" end
	if greenCan > .7 then return "MEDIUM" end
	if redCan > .7 then return "HIGH" end
end

local function reactionByColor(object)																											
	local redCan, greenCan, blueCan, alphaCan = object:GetStatusBarColor()
	if redCan < .01 and blueCan < .01 and greenCan > .99 then return "FRIENDLY", "NPC" 
	elseif redCan < .01 and blueCan > .99 and greenCan < .01 then return "FRIENDLY", "PLAYER"
	elseif redCan > .99 and blueCan < .01 and greenCan > .99 then return "NEUTRAL", "NPC"
	else return "HOSTILE", "PLAYER" end
end

local function GetName(LibNP)
	return LibNP.unit.name
end

function lib:GetName(frame)
	return self.nameplates[frame] and self.nameplates[frame].unit.name
end

local function GetLevel(LibNP)
	return LibNP.unit.level
end

function lib:GetLevel(frame)
	return self.nameplates[frame] and self.nameplates[frame].unit.level
end

local function GetReaction(LibNP)
	return reactionByColor(LibNP.bars.health)
end

function lib:GetReaction(frame)
	return self.nameplates[frame] and reactionByColor(self.nameplates[frame].bars.health)
end

local function GetType(LibNP)
	return select(2, reactionByColor(LibNP.bars.health))
end

function lib:GetType(frame)
	return self.nameplates[frame] and select(2, reactionByColor(self.nameplates[frame].bars.health))
end

local function IsBoss(LibNP)
	return LibNP.unit.isBoss
end

function lib:IsBoss(frame)
	return self.nameplates[frame] and self.nameplates[frame].unit.isBoss
end

local function GetHealthMax(LibNP)
	return LibNP.unit.healthmax
end

function lib:GetHealthMax(frame)
	return self.nameplates[frame] and self.nameplates[frame].unit.healthmax
end

local function GetClass(LibNP)
	return LibNP.unit.class
end

function lib:GetClass(frame)
	return self.nameplates[frame] and self.nameplates[frame].unit.class
end

local function IsElite(LibNP)
	return LibNP.unit.isElite
end

function lib:IsElite(frame)
	return self.nameplates[frame] and self.nameplates[frame].unit.isElite
end

local function GetThreatSituation(LibNP)
	return threatByColor(LibNP.regions.threatGlow)
end

function lib:GetThreatSituation(frame)
	return self.nameplates[frame] and threatByColor(self.nameplates[frame].regions.threatGlow)
end

local function GetAlpha(LibNP)
	return LibNP.frame:GetAlpha()
end

function lib:GetAlpha(frame)-- =/
	return frame:GetAlpha()
end

local function IsTarget(LibNP)
	return (LibNP.frame:GetAlpha() == 1 and UnitExists_("target"))
end

function lib:IsTarget(frame)
	return (frame:GetAlpha() == 1 and UnitExists_("target"))
end

local function GetHealth(LibNP)
	return LibNP.bars.health:GetValue()
end

function lib:GetHealth(frame)
	return self.nameplates[frame] and self.nameplates[frame].bars.health:GetValue()
end

local function IsMouseover(LibNP)
	return LibNP.regions.highlightTexture:IsShown()
end

function lib:IsMouseover(frame)
	return self.nameplates[frame] and self.nameplates[frame].regions.highlightTexture:IsShown()
end

local function IsCasting(LibNP)
	return LibNP.bars.cast:IsShown()
end

function lib:IsCasting(frame)
	return self.nameplates[frame] and self.nameplates[frame].bars.cast:IsShown()
end

local function IsInCombat(LibNP)
	return combatByColor(LibNP.regions.nameText:GetTextColor())
end

function lib:IsInCombat(frame)
	return self.nameplates[frame] and combatByColor(self.nameplates[frame].regions.nameText:GetTextColor())
end

local function IsMarked(LibNP)
	return LibNP.regions.raidIcon:IsShown()
end

function lib:IsMarked(frame)
	return self.nameplates[frame] and self.nameplates[frame].regions.raidIcon:IsShown()
end

local function GetRaidIcon(LibNP)
--~ 	return LibNP.unit.raidIcon or 0
	if LibNP.regions.raidIcon:IsShown() then
		local ULx,ULy = LibNP.regions.raidIcon:GetTexCoord() --,LLx,LLy,URx,URy,LRx,LRy
		if ULx and ULy then
			return raidIcons[ULx] and raidIcons[ULx][ULy]
		end
	end
end

function lib:GetRaidIcon(frame)
	if self.nameplates[frame] then
		if self.nameplates[frame].regions.raidIcon:IsShown() then
			local ULx,ULy = self.nameplates[frame].regions.raidIcon:GetTexCoord() --,LLx,LLy,URx,URy,LRx,LRy
			if ULx and ULy then
				return raidIcons[ULx] and raidIcons[ULx][ULy]
			end
		end
	end
	return nil
end

local function GetGUID(LibNP)
	return LibNP.unit.GUID
end

function lib:GetGUID(frame)
	return self.nameplates[frame] and self.nameplates[frame].unit.GUID
end

function LN_TEST()--	/run LN_TEST()
	debugPrint("Target", lib:GetTargetNameplate())
	local GUID = UnitGUID("target")
	debugPrint("GUID", lib:GetNameplateByGUID(GUID))
	local name = UnitName("target")
	debugPrint("Name", lib:GetNameplateByName(name))
end

-- ***
function lib:GetTargetNameplate()
	if self.targeted and self.nameplates[self.targeted] and self.nameplates[self.targeted].active == true then
		return self.targeted
	end
end

function lib:GetNameplateByGUID(GUID)
	if self.GUIDs[GUID] and self.nameplates[self.GUIDs[GUID]] and self.nameplates[self.GUIDs[GUID]].active == true then
		return self.GUIDs[GUID]
	end
end


function lib:GetNameplateByName(name)
	if self.names[name] and self.nameplates[self.names[name]] and self.nameplates[self.names[name]].active == true then
		return self.names[name]
	end
end

function lib:GetNameplateByUnit(unitID)
	if UnitIsUnit_(unitID, "target") then
		return self:GetTargetNameplate()
	end
	local GUID = UnitGUID_(unitID)
	if self.GUIDs[GUID] then
		return self.GUIDs[GUID], self.GUIDs[GUID].LibNP
	end
	local name = UnitName_(unitID)
	if self.names[GUID] then
		return self.names[GUID], self.names[GUID].LibNP
	end
end

function lib:GetPlateInfo(frame)
	return self.nameplates[frame]
end

local function destroyTable(t)
	setmetatable(t, nil)
	for k,v in pairs_(t) do
			t[k] = nil 
	end
end


local function RecycleNameplate(frame)
	local LibNP = lib.nameplates[frame]
	lib.callbacks:Fire("LibNameplate_RecycleNameplate", frame)

	if LibNP then
		local GUID = LibNP.unit.GUID
		local name = LibNP.unit.name
		local raidNum = LibNP.unit.raidIcon

		if lib.GUIDs[GUID] then lib.GUIDs[GUID] = nil end
		if lib.names[name] then lib.names[name] = nil end
	end

--~ 	lib.nameplates[frame].unit.name = nil
	lib.nameplates[frame].unit.GUID = nil
	lib.nameplates[frame].active = false
end




local function combatByColor(r, g, b, a) 
	return (r > .5 and g < .5)
end

--[[
	local threatGlow, healthBorder, castBorder, castNostop, spellIcon, highlightTexture, nameText, levelText, dangerSkull, raidIcon, eliteIcon
]]

local function GetMouseoverGUID(frame)
	local LibNP = lib.nameplates[frame]
	local GUID
	local unitID = "mouseover"
	if UnitExists_(unitID) then
		GUID = UnitGUID_(unitID)
		LibNP.unit.GUID = GUID
		lib.GUIDs[GUID] = frame
		
		
--~ 		local _, _, _, _, _, highlightTexture, nameTextRegion = frame:GetRegions()
--~ 		debugPrint("Found Mouseover GUID", UnitName_(unitID), GUID, UnitName_(unitID) == nameTextRegion:GetText())
		
		lib.callbacks:Fire("LibNameplate_FoundGUID", frame, GUID, "mouseover")
	end
end

local function FindPlateWithRaidIcon(iconNum, unitID)
	local GUID = UnitGUID_(unitID)
	local LibNP
	for frame, LibNP in pairs_(lib.nameplates) do 
		if lib:IsMarked(frame) and lib:GetRaidIcon(frame) == iconNum then
			return frame, LibNP
		end
	end
	return nil
end

local function FindGUIDByRaidIcon(frame, raidNum)
	local LibNP = lib.nameplates[frame]
	local group, num = "", 0
	if GetNumRaidMembers_() > 1 then
		group, num = "raid", GetNumRaidMembers_()
	elseif GetNumPartyMembers_() > 0 then
		group, num = "party", GetNumPartyMembers_()
	end
	
	local unitID, targetIndex, GUID
	for i = 1, num do
		unitID = "party"..i;
		if not UnitIsUnit_("target", unitID.."target") then
			targetIndex = GetRaidTargetIndex_(unitID.."target")
			if targetIndex and targetIndex == raidNum then
				GUID = UnitGUID_(unitID.."target")
				LibNP.unit.GUID = GUID
				lib.GUIDs[GUID] = frame
				lib.callbacks:Fire("LibNameplate_FoundGUID", frame, GUID, unitID.."target")
				return
			end
		end
	end
end

local function UpdateNameplateInfo(frame)
	local LibNP = lib.nameplates[frame]
	local unit = LibNP.unit
	local regions = LibNP.regions

--~ 	debugPrint("UpdateNameplateInfo", unit.name)
	
	if not unit.GUID and regions.highlightTexture:IsShown() then --mouseover
		GetMouseoverGUID(frame)
	end
	
	--So I don't forget, we keep checking the raid icon so try to find the GUID. 
	if not unit.GUID then
		local raidNum = lib:GetRaidIcon(frame)
		if raidNum and raidNum > 0 then
			FindGUIDByRaidIcon(frame, raidNum)
		end
	end
	
--~ 	debugPrint("Updated nameplate ",unit.name)
	frame.lnpLastUpdate = 0
end

local onShowHooks = {}
local onHideHooks = {}
local onUpdateHooks = {}

local function nameplateOnUpdate(self, elapsed, ...)
	if self.lnpCheckForTarget == true then 
		self.lnpCheckForTarget = false
		if not lib.nameplates[self].unit.GUID and self:IsShown() and ((self:GetAlpha() == 1) and UnitExists_("target")) then
			local GUID = UnitGUID_("target")
			lib.nameplates[self].unit.GUID = GUID
			lib.GUIDs[GUID] = self
			lib.callbacks:Fire("LibNameplate_FoundGUID", self, GUID, "target")
		end
	end
	
	self.lnpLastUpdate = self.lnpLastUpdate + elapsed
	if self.lnpLastUpdate > updateDelay then
		UpdateNameplateInfo(self, true)
	end

	if onUpdateHooks[self] and onUpdateHooks[self] ~= nameplateOnUpdate then
		onUpdateHooks[self](self, elapsed, ...)
	end
end

local function nameplateOnHide(self, ...)
--~ 	debugPrint("OnHide!")
	RecycleNameplate(self)
	
	--Onhook the OnUpdate. Thought since OnUpdate doesn't fire when hidden this may not be needed.
	if onUpdateHooks[self] then
		self:SetScript("OnUpdate", nil)
		onUpdateHooks[self] = nil
	end

	if onHideHooks[self] and onHideHooks[self] ~= nameplateOnHide then
		onHideHooks[self](self, ...)
	end
end


local function nameplateOnShow(self, ...)
--~ 	debugPrint("OnShow!")
	lib:SetupNameplate(self)
	if onShowHooks[self] and onShowHooks[self] ~= nameplateOnShow then
		onShowHooks[self](self, ...)
	end
end

function lib:SetupNameplate(frame)
	if not lib.nameplates[frame] then
		lib.nameplates[frame] = {}
		local LibNP = lib.nameplates[frame]
		LibNP.frame = frame
		
		--REGIONS
		if not LibNP.regions then
			local regs = {frame:GetRegions()}
			LibNP.regions = {}
			
			for i, name in pairs_(regionNames) do 
				LibNP.regions[name] = regs[i]
			end
		end
		local regions = LibNP.regions
		
		--BARS
		LibNP.bars = LibNP.bars or {}
		local bars = LibNP.bars
		bars.health, bars.cast = frame:GetChildren()
		
		--UNIT
		LibNP.unit = LibNP.unit or {}
		
		--FUNC
		LibNP.GetName				= GetName
		LibNP.GetReaction			= GetReaction
		LibNP.GetHealthMax			= GetHealthMax
		LibNP.GetType				= GetType
		LibNP.GetLevel				= GetLevel
		LibNP.GetClass				= GetClass
		LibNP.IsElite				= IsElite
		LibNP.IsBoss				= IsBoss
		LibNP.GetThreatSituation	= GetThreatSituation
		LibNP.GetAlpha				= GetAlpha
		LibNP.IsTarget				= IsTarget
		LibNP.GetHealth				= GetHealth
		LibNP.IsMouseover			= IsMouseover
		LibNP.IsMarked				= IsMarked
		LibNP.GetRaidIcon			= GetRaidIcon
		LibNP.GetGUID				= GetGUID
		
		if frame:HasScript("OnHide") and frame:GetScript("OnHide") ~= nameplateOnHide then
			onHideHooks[frame] = frame:GetScript("OnHide")
			frame:SetScript("OnHide", nameplateOnHide)
		end

		if frame:HasScript("OnShow") and frame:GetScript("OnShow") ~= nameplateOnShow then
			onShowHooks[frame] = frame:GetScript("OnShow")
			frame:SetScript("OnShow", nameplateOnShow)
		end

		frame.lnpLastUpdate = 0
	end
	
	--Hook the OnUpdate script. There we check if a raid icon has appeared and we attempt to find the GUID from that icon.
	if frame:HasScript("OnUpdate") and frame:GetScript("OnUpdate") ~= nameplateOnUpdate then
		onUpdateHooks[frame] = frame:GetScript("OnUpdate")
		frame:SetScript("OnUpdate", nameplateOnUpdate) 
	end
	
	
	
	--Info that only needs to be pulled once. Name, level, isBoss ect.
	local LibNP = lib.nameplates[frame]
	local regions = LibNP.regions
	local plateName = string_gsub(regions.nameText:GetText(), "|cffffffff", "") --Sometimes the plate name is coloured white.
	
	local unit = LibNP.unit
	unit.name = plateName
	lib.names[plateName] = frame
	unit.isBoss = regions.dangerSkull:IsShown()
	unit.level = regions.levelText:GetText()
	
	local bars = LibNP.bars
	unit.red, unit.green, unit.blue = bars.health:GetStatusBarColor()
	_, unit.healthmax = bars.health:GetMinMaxValues()
	unit.class = colorToClass["C"..pctToInt(unit.red)+pctToInt(unit.green)+pctToInt(unit.blue)] or "UNKNOWN"
	unit.isElite = regions.eliteIcon:IsShown()
	
	
	LibNP.active = true
	lib.callbacks:Fire("LibNameplate_NewNameplate", frame)
--~ 	--~ 		debugPrint("LibNameplate_NewNameplate", LibNP:GetName())
	frame.lnpCheckForTarget = true
	UpdateNameplateInfo(frame)
end--	/lnbug


local function FindNameplates(...)
--~ 	debugPrint("FindNameplates ")
	for i = 1, select_('#', ...) do
		local frame = select_(i, ...)
		if frame:IsShown() then
			if not lib.nameplates[frame] and IsNamePlateFrame(frame) then
				lib:SetupNameplate(frame, true)
			end
		end
	end
end

local function CheckForTargetGUID()
	local unitID = "target"
	local GUID
	local highlightTexture, nameTextRegion
	for frame, LibNP in pairs_(lib.nameplates) do 
	
		if frame:IsShown() and frame:GetAlpha() == 1 and UnitExists_(unitID) then
			lib.targeted = frame
			if not LibNP.unit.GUID then
				GUID = UnitGUID_(unitID)
				LibNP.unit.GUID = GUID
				lib.GUIDs[GUID] = frame
				lib.callbacks:Fire("LibNameplate_FoundGUID", frame, GUID, unitID)
				
--~ 				_, _, _, _, _, highlightTexture, nameTextRegion = frame:GetRegions()
--~ 				debugPrint("Found target GUID", UnitName_(unitID), GUID, UnitName_(unitID) == nameTextRegion:GetText())
			end
			lib.callbacks:Fire("LibNameplate_TargetNameplate", frame)
			return
		end
	end
--~ 	debugPrint("Failed to find target GUID")
end

local function OnUpdate(self, elapsed)
	if self.checkTarget == true then
		self.checkTarget = false
		CheckForTargetGUID()

	end
	self.lastUpdate = self.lastUpdate + elapsed
	if self.lastUpdate >= scanDelay then
		self.lastUpdate = 0
		FindNameplates(WorldFrame_:GetChildren())
	end
end


local function OnEvent(self, event, ...)
	if event == "UPDATE_MOUSEOVER_UNIT" then
		if GetMouseFocus_():GetName() == "WorldFrame" then
			for frame, LibNP in pairs_(lib.nameplates) do 
				if LibNP.regions.highlightTexture:IsShown() then
					if not LibNP.unit.GUID then
						GetMouseoverGUID(frame)
					end
					lib.callbacks:Fire("LibNameplate_MouseoverNameplate", frame)
					return
				end
			end
		end
		
	elseif event == "PLAYER_TARGET_CHANGED" then
		if UnitExists_("target") then
			self.checkTarget = true--Check for target nameplate on next OnUpdate. We can't do it now.
		else
			lib.targeted = nil
		end
		
	elseif event == "UNIT_TARGET" then
		local unitID = ...
		local targetID = unitID.."target"
		if not UnitIsUnit("player", unitID) then
			local iconNum = GetRaidTargetIndex_(targetID)
			if iconNum and iconNum > 0 then
				local foundPlate = FindPlateWithRaidIcon(iconNum, targetID)
				local LibNP = lib.nameplates[foundPlate]
				if foundPlate and not LibNP.unit.GUID then
					local GUID = UnitGUID_(targetID)
					LibNP.unit.GUID = GUID
					lib.GUIDs[GUID] = frame
					lib.callbacks:Fire("LibNameplate_FoundGUID", frame, GUID, targetID)
				end
			end
		end
		
	elseif event == "RAID_TARGET_UPDATE" then
		local regions
		for frame, LibNP in pairs_(lib.nameplates) do 
			if frame:IsShown() and not LibNP.unit.GUID and lib:IsMarked(frame) then
				local raidNum = lib:GetRaidIcon(frame)
				if raidNum and raidNum > 0 then
					FindGUIDByRaidIcon(frame, raidNum)
				end
			end
		end
	end
end


if not lib.frame then
	lib.frame=CreateFrame("Frame")
end

lib.frame.lastUpdate = 0
lib.frame.checkTarget = false
lib.frame:SetScript("OnUpdate", OnUpdate)
lib.frame:SetScript("OnEvent", OnEvent)
lib.frame:RegisterEvent('UPDATE_MOUSEOVER_UNIT')
lib.frame:RegisterEvent('PLAYER_TARGET_CHANGED')
lib.frame:RegisterEvent('UNIT_TARGET')
lib.frame:RegisterEvent('RAID_TARGET_UPDATE')

